@@ -25,8 +25,8 @@
 #include "eval.h"
 #include "exec.h"
 #include "printf.h"
+#include "regex.h"
 #include "stat.h"
-#include <regex.h>
 #include <stdbool.h>
 #include <stddef.h>
 #include <sys/types.h>
@@ -159,7 +159,7 @@ struct expr {
 	CFILE *cfile;
 
 	/** Optional compiled regex. */
-	regex_t *regex;
+	struct bfs_regex *regex;
 
 	/** Optional exec command. */
 	struct bfs_exec *execbuf;
