@@ -2,7 +2,7 @@
 
 ############################################################################
 # bfs                                                                      #
-# Copyright (C) 2015-2021 Tavian Barnes <tavianator@tavianator.com>        #
+# Copyright (C) 2015-2022 Tavian Barnes <tavianator@tavianator.com>        #
 #                                                                          #
 # Permission to use, copy, modify, and/or distribute this software for any #
 # purpose with or without fee is hereby granted.                           #
@@ -16,8 +16,7 @@
 # OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           #
 ############################################################################
 
-set -e
-set -o physical
+set -eP
 umask 022
 
 export LC_ALL=C
@@ -40,6 +39,8 @@ if [ -t 1 ]; then
     RST="$(printf '\033[0m')"
 fi
 
+UNAME="$(uname)"
+
 if command -v capsh &>/dev/null; then
     if capsh --has-p=cap_dac_override &>/dev/null || capsh --has-p=cap_dac_read_search &>/dev/null; then
 	if [ -n "$BFS_TRIED_DROP" ]; then
@@ -63,7 +64,7 @@ EOF
     fi
 elif [ "$EUID" -eq 0 ]; then
     UNLESS=
-    if [ "$(uname)" = "Linux" ]; then
+    if [ "$UNAME" = "Linux" ]; then
 	UNLESS=" unless ${GRN}capsh${RST} is installed"
     fi
 
@@ -107,17 +108,6 @@ Usage: ${GRN}$0${RST} [${BLU}--bfs${RST}=${MAG}path/to/bfs${RST}] [${BLU}--posix
 EOF
 }
 
-function _realpath() {
-    (
-        cd "${1%/*}"
-        echo "$PWD/${1##*/}"
-    )
-}
-
-BFS="$(_realpath ./bfs)"
-TESTS="$(_realpath ./tests)"
-UNAME="$(uname)"
-
 DEFAULT=yes
 POSIX=
 BSD=
@@ -242,6 +232,11 @@ posix_tests=(
     test_name_root
     test_name_root_depth
     test_name_trailing_slash
+    test_name_star_star
+    test_name_character_class
+    test_name_bracket
+    test_name_backslash
+    test_name_double_backslash
 
     test_newer
     test_newer_link
@@ -271,6 +266,7 @@ posix_tests=(
     test_perm_sticky
 
     test_prune
+    test_prune_file
     test_prune_or_print
     test_not_prune
 
@@ -362,6 +358,7 @@ bsd_tests=(
     test_execdir_ulimit
 
     test_exit
+    test_exit_no_implicit_print
 
     test_flags
 
@@ -495,6 +492,10 @@ gnu_tests=(
 
     test_exec_nothing
     test_exec_substring
+    test_exec_flush
+    test_exec_flush_fail
+    test_exec_plus_flush
+    test_exec_plus_flush_fail
 
     test_execdir
     test_execdir_substring
@@ -558,6 +559,8 @@ gnu_tests=(
 
     test_iregex
 
+    test_iwholename
+
     test_lname
     test_L_lname
 
@@ -625,6 +628,10 @@ gnu_tests=(
 
     test_regextype_posix_basic
     test_regextype_posix_extended
+    test_regextype_ed
+    test_regextype_emacs
+    test_regextype_grep
+    test_regextype_sed
 
     test_samefile
     test_samefile_symlink
@@ -647,6 +654,8 @@ gnu_tests=(
     test_uid_minus
     test_uid_minus_plus
 
+    test_wholename
+
     test_writable
 
     test_xtype_l
@@ -735,6 +744,9 @@ bfs_tests=(
     test_color_star
     test_color_ls
 
+    test_exec_flush_fprint
+    test_exec_flush_fprint_fail
+
     test_execdir_plus
 
     test_fprint_duplicate_stdout
@@ -851,6 +863,23 @@ fi
 
 eval enabled_tests=($(printf '%q\n' "${enabled_tests[@]}" | sort -u))
 
+function _realpath() {
+    (
+        cd "$(dirname -- "$1")"
+        echo "$PWD/$(basename -- "$1")"
+    )
+}
+
+ROOT="$(dirname -- "${BASH_SOURCE[0]}")"
+
+# Try to resolve the path to $BFS before we cd, while also supporting
+# --bfs="./bfs -S ids"
+BFS=(${BFS:-$ROOT/bfs})
+BFS[0]="$(_realpath "$(command -v "${BFS[0]}")")"
+BFS="${BFS[*]}"
+
+TESTS="$(_realpath "$ROOT/tests")"
+
 # The temporary directory that will hold our test data
 TMP="$(mktemp -d "${TMPDIR:-/tmp}"/bfs.XXXXXXXXXX)"
 chown "$(id -u):$(id -g)" "$TMP"
@@ -978,6 +1007,7 @@ function make_weirdnames() {
     touchp "$1/.../h"
     touchp "$1/\\/i"
     touchp "$1/ /j"
+    touchp "$1/[/k"
 }
 make_weirdnames "$TMP/weirdnames"
 
@@ -1108,10 +1138,19 @@ function fail() {
     fi
 }
 
+# Detect colored diff support
+if diff --color /dev/null /dev/null 2>/dev/null; then
+    DIFF="diff --color"
+else
+    DIFF="diff"
+fi
+
 # Return value when bfs fails
 EX_BFS=10
 # Return value when a difference is detected
 EX_DIFF=20
+# Return value when a test is skipped
+EX_SKIP=77
 
 function bfs_diff() (
     bfs_verbose "$@"
@@ -1138,7 +1177,7 @@ function bfs_diff() (
     local STATUS="${PIPESTATUS[0]}"
 
     if [ ! "$UPDATE" ]; then
-        diff -u "$EXPECTED" "$ACTUAL" || return $EX_DIFF
+        $DIFF -u "$EXPECTED" "$ACTUAL" || return $EX_DIFF
     fi
 
     if [ "$STATUS" -eq 0 ]; then
@@ -1148,6 +1187,12 @@ function bfs_diff() (
     fi
 )
 
+function skip_if() {
+    if "$@"; then
+        exit $EX_SKIP
+    fi
+}
+
 function closefrom() {
     if [ -d /proc/self/fd ]; then
         local fds=/proc/self/fd
@@ -1275,10 +1320,43 @@ function test_name_slashes() {
     bfs_diff /// -maxdepth 0 -name /
 }
 
+function test_name_star_star() {
+    bfs_diff basic -name '**f**'
+}
+
+function test_name_character_class() {
+    bfs_diff basic -name '[e-g][!a-n][!p-z]'
+}
+
+function test_name_bracket() {
+    # fnmatch() is broken on macOS
+    skip_if test "$UNAME" = "Darwin"
+
+    # An unclosed [ should be matched literally
+    bfs_diff weirdnames -name '['
+}
+
+function test_name_backslash() {
+    # fnmatch() is broken on macOS
+    skip_if test "$UNAME" = "Darwin"
+
+    # An unescaped \ doesn't match
+    bfs_diff weirdnames -name '\'
+}
+
+function test_name_double_backslash() {
+    # An escaped \\ matches
+    bfs_diff weirdnames -name '\\'
+}
+
 function test_path() {
     bfs_diff basic -path 'basic/*f*'
 }
 
+function test_wholename() {
+    bfs_diff basic -wholename 'basic/*f*'
+}
+
 function test_true() {
     bfs_diff basic -true
 }
@@ -1541,18 +1619,26 @@ function test_xtype_depth() {
 }
 
 function test_iname() {
+    skip_if fail quiet invoke_bfs -quit -iname PATTERN
     bfs_diff basic -iname '*F*'
 }
 
 function test_ipath() {
+    skip_if fail quiet invoke_bfs -quit -ipath PATTERN
     bfs_diff basic -ipath 'basic/*F*'
 }
 
+function test_iwholename() {
+    skip_if fail quiet invoke_bfs -quit -iwholename PATTERN
+    bfs_diff basic -iwholename 'basic/*F*'
+}
+
 function test_lname() {
     bfs_diff links -lname '[aq]'
 }
 
 function test_ilname() {
+    skip_if fail quiet invoke_bfs -quit -ilname PATTERN
     bfs_diff links -ilname '[AQ]'
 }
 
@@ -1561,6 +1647,7 @@ function test_L_lname() {
 }
 
 function test_L_ilname() {
+    skip_if fail quiet invoke_bfs -quit -ilname PATTERN
     bfs_diff -L links -ilname '[AQ]'
 }
 
@@ -1674,6 +1761,36 @@ function test_exec_substring() {
     bfs_diff basic -exec echo '-{}-' ';'
 }
 
+function test_exec_flush() {
+    # IO streams should be flushed before executing programs
+    bfs_diff basic -printf '%p ' -exec echo found \;
+}
+
+function test_exec_flush_fail() {
+    # Failure to flush streams before exec should be caught
+    skip_if test ! -e /dev/full
+    fail quiet invoke_bfs basic -printf '%p ' -exec true \; >/dev/full
+}
+
+function test_exec_flush_fprint() {
+    # Even non-stdstreams should be flushed
+    bfs_diff basic/a -fprint scratch/foo -exec cat scratch/foo \;
+}
+
+function test_exec_flush_fprint_fail() {
+    skip_if test ! -e /dev/full
+    fail quiet invoke_bfs basic/a -fprint /dev/full -exec true \;
+}
+
+function test_exec_plus_flush() {
+    bfs_diff basic/a -printf '%p ' -exec echo found {} +
+}
+
+function test_exec_plus_flush_fail() {
+    skip_if test ! -e /dev/full
+    fail quiet invoke_bfs basic/a -printf '%p ' -exec echo found {} + >/dev/full
+}
+
 function test_execdir() {
     bfs_diff basic -execdir echo '{}' ';'
 }
@@ -1750,7 +1867,7 @@ function test_fprint() {
     if [ "$UPDATE" ]; then
         cp {scratch,"$TESTS"}/test_fprint.out
     else
-        diff -u {"$TESTS",scratch}/test_fprint.out
+        $DIFF -u {"$TESTS",scratch}/test_fprint.out
     fi
 }
 
@@ -1765,7 +1882,7 @@ function test_fprint_duplicate() {
     if [ "$UPDATE" ]; then
         cp {scratch,"$TESTS"}/test_fprint_duplicate.out
     else
-        diff -u {"$TESTS",scratch}/test_fprint_duplicate.out
+        $DIFF -u {"$TESTS",scratch}/test_fprint_duplicate.out
     fi
 }
 
@@ -1778,7 +1895,7 @@ function test_fprint_duplicate_stdout() {
     if [ "$UPDATE" ]; then
         cp {scratch,"$TESTS"}/test_fprint_duplicate_stdout.out
     else
-        diff -u {"$TESTS",scratch}/test_fprint_duplicate_stdout.out
+        $DIFF -u {"$TESTS",scratch}/test_fprint_duplicate_stdout.out
     fi
 }
 
@@ -1799,7 +1916,7 @@ function test_fprint_truncate() {
     if [ "$UPDATE" ]; then
         cp {scratch,"$TESTS"}/test_fprint_truncate.out
     else
-        diff -u {"$TESTS",scratch}/test_fprint_truncate.out
+        $DIFF -u {"$TESTS",scratch}/test_fprint_truncate.out
     fi
 }
 
@@ -1934,6 +2051,10 @@ function test_prune() {
     bfs_diff basic -name foo -prune
 }
 
+function test_prune_file() {
+    bfs_diff basic -print -name '?' -prune
+}
+
 function test_prune_or_print() {
     bfs_diff basic -name foo -prune -o -print
 }
@@ -2040,6 +2161,28 @@ function test_regextype_posix_extended() {
     bfs_diff -regextype posix-extended -regex '\./(\()'
 }
 
+function test_regextype_ed() {
+    cd weirdnames
+    bfs_diff -regextype ed -regex '\./\((\)'
+}
+
+function test_regextype_emacs() {
+    skip_if fail quiet invoke_bfs -regextype emacs -quit
+
+    bfs_diff basic -regextype emacs -regex '.*/\(f+o?o?\|bar\)'
+}
+
+function test_regextype_grep() {
+    skip_if fail quiet invoke_bfs -regextype grep -quit
+
+    bfs_diff basic -regextype grep -regex '.*/f\+o\?o\?'
+}
+
+function test_regextype_sed() {
+    cd weirdnames
+    bfs_diff -regextype sed -regex '\./\((\)'
+}
+
 function test_d_path() {
     bfs_diff -d basic
 }
@@ -2059,7 +2202,7 @@ function test_s() {
     if [ "$UPDATE" ]; then
         cp {"$TMP","$TESTS"}/test_s.out
     else
-        diff -u {"$TESTS","$TMP"}/test_s.out
+        $DIFF -u {"$TESTS","$TMP"}/test_s.out
     fi
 }
 
@@ -2257,7 +2400,7 @@ function test_printf_nul() {
     invoke_bfs basic -maxdepth 0 -printf '%h\0%f\n' >"$ACTUAL"
 
     if [ ! "$UPDATE" ]; then
-        diff -u "$EXPECTED" "$ACTUAL"
+        $DIFF -u "$EXPECTED" "$ACTUAL"
     fi
 }
 
@@ -2330,7 +2473,7 @@ function test_fprintf() {
     if [ "$UPDATE" ]; then
         cp scratch/test_fprintf.out "$TESTS/test_fprintf.out"
     else
-        diff -u "$TESTS/test_fprintf.out" scratch/test_fprintf.out
+        $DIFF -u "$TESTS/test_fprintf.out" scratch/test_fprintf.out
     fi
 }
 
@@ -2446,7 +2589,7 @@ function test_color_nul() {
     LS_COLORS="ec=\33[m\0:" invoke_bfs rainbow -color -maxdepth 0 >"$ACTUAL"
 
     if [ ! "$UPDATE" ]; then
-        diff -u "$EXPECTED" "$ACTUAL"
+        $DIFF -u "$EXPECTED" "$ACTUAL"
     fi
 }
 
@@ -2586,7 +2729,7 @@ function test_color_ls() {
         | sort -o "$ACTUAL"
 
     if [ ! "$UPDATE" ]; then
-        diff -u "$EXPECTED" "$ACTUAL"
+        $DIFF -u "$EXPECTED" "$ACTUAL"
     fi
 }
 
@@ -2619,6 +2762,10 @@ function test_exit() {
     bfs_diff basic/g -print -name g -exit
 }
 
+function test_exit_no_implicit_print() {
+    bfs_diff basic -not -name foo -o -exit
+}
+
 function test_printx() {
     bfs_diff weirdnames -printx
 }
@@ -2706,35 +2853,29 @@ function test_data_flow_or_swap() {
 }
 
 function test_print_error() {
-    if [ -e /dev/full ]; then
-        fail quiet invoke_bfs basic -maxdepth 0 >/dev/full
-    fi
+    skip_if test ! -e /dev/full
+    fail quiet invoke_bfs basic -maxdepth 0 >/dev/full
 }
 
 function test_fprint_error() {
-    if [ -e /dev/full ]; then
-        fail quiet invoke_bfs basic -maxdepth 0 -fprint /dev/full
-    fi
+    skip_if test ! -e /dev/full
+    fail quiet invoke_bfs basic -maxdepth 0 -fprint /dev/full
 }
 
 function test_fprint_noerror() {
     # Regression test: /dev/full should not fail until actually written to
-
-    if [ -e /dev/full ]; then
-        invoke_bfs basic -false -fprint /dev/full
-    fi
+    skip_if test ! -e /dev/full
+    invoke_bfs basic -false -fprint /dev/full
 }
 
 function test_fprint_error_stdout() {
-    if [ -e /dev/full ]; then
-        fail quiet invoke_bfs basic -maxdepth 0 -fprint /dev/full >/dev/full
-    fi
+    skip_if test ! -e /dev/full
+    fail quiet invoke_bfs basic -maxdepth 0 -fprint /dev/full >/dev/full
 }
 
 function test_fprint_error_stderr() {
-    if [ -e /dev/full ]; then
-        fail invoke_bfs basic -maxdepth 0 -fprint /dev/full 2>/dev/full
-    fi
+    skip_if test ! -e /dev/full
+    fail invoke_bfs basic -maxdepth 0 -fprint /dev/full 2>/dev/full
 }
 
 function test_print0() {
@@ -2923,10 +3064,10 @@ function set_acl() {
 function test_acl() {
     rm -rf scratch/*
 
-    quiet invoke_bfs scratch -quit -acl || return 0
+    skip_if fail quiet invoke_bfs scratch -quit -acl
 
     $TOUCH scratch/{normal,acl}
-    set_acl scratch/acl || return 0
+    skip_if fail set_acl scratch/acl
     ln -s acl scratch/link
 
     bfs_diff scratch -acl
@@ -2935,10 +3076,10 @@ function test_acl() {
 function test_L_acl() {
     rm -rf scratch/*
 
-    quiet invoke_bfs scratch -quit -acl || return 0
+    skip_if fail quiet invoke_bfs scratch -quit -acl
 
     $TOUCH scratch/{normal,acl}
-    set_acl scratch/acl || return 0
+    skip_if fail set_acl scratch/acl
     ln -s acl scratch/link
 
     bfs_diff -L scratch -acl
@@ -2947,9 +3088,7 @@ function test_L_acl() {
 function test_capable() {
     rm -rf scratch/*
 
-    if fail quiet invoke_bfs scratch -quit -capable; then
-        return 0
-    fi
+    skip_if fail quiet invoke_bfs scratch -quit -capable
 
     $TOUCH scratch/{normal,capable}
     sudo setcap all+ep scratch/capable
@@ -2961,9 +3100,7 @@ function test_capable() {
 function test_L_capable() {
     rm -rf scratch/*
 
-    if fail quiet invoke_bfs scratch -quit -capable; then
-        return 0
-    fi
+    skip_if fail quiet invoke_bfs scratch -quit -capable
 
     $TOUCH scratch/{normal,capable}
     sudo setcap all+ep scratch/capable
@@ -3001,20 +3138,20 @@ function make_xattrs() {
 }
 
 function test_xattr() {
-    quiet invoke_bfs scratch -quit -xattr || return 0
-    make_xattrs || return 0
+    skip_if fail quiet invoke_bfs scratch -quit -xattr
+    skip_if fail make_xattrs
     bfs_diff scratch -xattr
 }
 
 function test_L_xattr() {
-    quiet invoke_bfs scratch -quit -xattr || return 0
-    make_xattrs || return 0
+    skip_if fail quiet invoke_bfs scratch -quit -xattr
+    skip_if fail make_xattrs
     bfs_diff -L scratch -xattr
 }
 
 function test_xattrname() {
-    quiet invoke_bfs scratch -quit -xattr || return 0
-    make_xattrs || return 0
+    skip_if fail quiet invoke_bfs scratch -quit -xattr
+    skip_if fail make_xattrs
 
     case "$UNAME" in
         Darwin|FreeBSD)
@@ -3027,8 +3164,8 @@ function test_xattrname() {
 }
 
 function test_L_xattrname() {
-    quiet invoke_bfs scratch -quit -xattr || return 0
-    make_xattrs || return 0
+    skip_if fail quiet invoke_bfs scratch -quit -xattr
+    skip_if fail make_xattrs
 
     case "$UNAME" in
         Darwin|FreeBSD)
@@ -3091,7 +3228,7 @@ function test_S() {
     if [ "$UPDATE" ]; then
         cp {"$TMP","$TESTS"}/"test_S_$1.out"
     else
-        diff -u {"$TESTS","$TMP"}/"test_S_$1.out"
+        $DIFF -u {"$TESTS","$TMP"}/"test_S_$1.out"
     fi
 }
 
@@ -3128,12 +3265,12 @@ function test_exclude_exclude() {
 }
 
 function test_flags() {
-    quiet invoke_bfs scratch -quit -flags offline || return 0
+    skip_if fail quiet invoke_bfs scratch -quit -flags offline
 
     rm -rf scratch/*
 
     $TOUCH scratch/{foo,bar}
-    chflags offline scratch/bar || return 0
+    skip_if fail chflags offline scratch/bar
 
     bfs_diff scratch -flags -offline,nohidden
 }
@@ -3170,15 +3307,13 @@ function test_files0_from_ok() {
 }
 
 function test_stderr_fails_silently() {
-    if [ -e /dev/full ]; then
-        bfs_diff -D all basic 2>/dev/full
-    fi
+    skip_if test ! -e /dev/full
+    bfs_diff -D all basic 2>/dev/full
 }
 
 function test_stderr_fails_loudly() {
-    if [ -e /dev/full ]; then
-        fail invoke_bfs -D all basic -false -fprint /dev/full 2>/dev/full
-    fi
+    skip_if test ! -e /dev/full
+    fail invoke_bfs -D all basic -false -fprint /dev/full 2>/dev/full
 }
 
 function test_unexpected_operator() {
@@ -3209,6 +3344,7 @@ fi
 
 passed=0
 failed=0
+skipped=0
 
 for test in "${enabled_tests[@]}"; do
     printf "${BOL}${YLW}%s${RST}${EOL}" "$test"
@@ -3218,6 +3354,11 @@ for test in "${enabled_tests[@]}"; do
 
     if [ $status -eq 0 ]; then
         ((++passed))
+    elif [ $status -eq $EX_SKIP ]; then
+        ((++skipped))
+        if [ "$VERBOSE" ]; then
+            printf "${BOL}${CYN}%s skipped!${RST}\n" "$test"
+        fi
     else
         ((++failed))
         printf "${BOL}${RED}%s failed!${RST}\n" "$test"
@@ -3227,6 +3368,9 @@ done
 if [ $passed -gt 0 ]; then
     printf "${BOL}${GRN}tests passed: %d${RST}\n" "$passed"
 fi
+if [ $skipped -gt 0 ]; then
+    printf "${BOL}${CYN}tests skipped: %s${RST}\n" "$skipped"
+fi
 if [ $failed -gt 0 ]; then
     printf "${BOL}${RED}tests failed: %s${RST}\n" "$failed"
     exit 1
